<?php

/**
 * @file
 * Handles webhooks from GoCardless.com.
 */

/**
 * Menu callback: respond to a webhook from GoCardless.com.
 *
 * @see commerce_gc_client_webhook_mandates()
 * @see commerce_gc_client_webhook_subscriptions()
 * @see commerce_gc_client_webhook_payments()
 */
function commerce_gc_client_webhook() {
  // Code for retrieving webhook headers for Nginx server.
  if (!function_exists('getallheaders')) {

    /**
     *
     */
    function getallheaders() {
      $headers = array();
      foreach ($_SERVER as $name => $value) {
        if (substr($name, 0, 5) == 'HTTP_') {
          $headers[str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5)))))] = $value;
        }
      }
      return $headers;
    }

  }

  // Code if hash_equals function is not present (PHP < 5.6.0)
  if (!function_exists('hash_equals')) {

    /**
     *
     */
    function hash_equals($known_string, $user_string) {
      $ret = 0;
      if (strlen($known_string) !== strlen($user_string)) {
        $user_string = $known_string;
        $ret = 1;
      }
      $res = $known_string ^ $user_string;
      for ($i = strlen($res) - 1; $i >= 0; --$i) {
        $ret |= ord($res[$i]);
      }
      return !$ret;
    }

  }

  $headers = getallheaders();
  $provided_signature = $headers["Webhook-Signature"];
  $calculated_signature = FALSE;
  $settings = CommerceGcClientPartner::getSettings();

  if ($settings['webhook']) {
    $secret = $settings['webhook'];
    $webhook = file_get_contents('php://input');
    $calculated_signature = hash_hmac("sha256", $webhook, $secret);
  }

  if (hash_equals($provided_signature, $calculated_signature)) {

    $data = json_decode($webhook, TRUE);

    // Optionally write webhook to log.
    if ($settings['log']['log_webhook']) {
      watchdog("commerce_gc_client", 'GoCardless webhook:<br /><pre>@data</pre>', array(
        '@data' => print_r($data, TRUE),
      ));
    }

    // Send a success header before processing webhooks, in order to
    // prevent timeouts.
    header('HTTP/1.1 200 OK');

    // Process the events.
    foreach ($data['events'] as $event) {

      switch ($event['resource_type']) {
        case 'mandates':
          $order_id = commerce_gc_client_get_order_id($event['links']['mandate']);
          $order = commerce_order_load($order_id);
          $resource = array();
          commerce_gc_client_webhook_mandates($order, $event['action'], $event);
          break;

        case 'subscriptions':
          $order_id = commerce_gc_client_get_order_id($event['links']['subscription']);
          $order = commerce_order_load($order_id);
          $resource = array();
          commerce_gc_client_webhook_subscriptions($order, $event['action'], $event);
          break;

        case 'payments':
          $payment_id = $event['links']['payment'];
          $partner = new CommerceGcClientPartner();
          $result = $partner->api([
            'endpoint' => 'payments',
            'action' => 'get',
            'id' => $payment_id,
          ]
          );
          if ($result->response->status_code == 200) {
            $resource = $result->response->body->payments;
            $order_id = commerce_gc_client_get_order_id($resource->links->mandate);
            $order = commerce_order_load($order_id);
            commerce_gc_client_webhook_payments($order, $event['action'], $event, $resource);
          }
          break;

        default:
          $order_id = FALSE;
          break;
      }

      // Provide hook so other modules can respond to GoCardless webhooks.
      module_invoke_all('commerce_gc_client_webhook', array(
        'event' => $event,
        'resource' => $resource,
        'order_id' => $order_id,
      ));
    }
  }
  else {
    watchdog("commerce_gc_client", 'GoCardless webhooks cannot be processed because the webhook secret is not set or is invalid.', array(), WATCHDOG_WARNING);
    header("HTTP/1.1 498 Invalid Token");
  }
}

/**
 * Processes 'mandates' webhooks.
 *
 * @param object $order
 *   Commerce order entity.
 * @param string $action
 *   The GoCardless action that triggerd the sending of the webhook.
 * @param array $event
 *   The webhook event as provided by GoCardless.
 *
 * @see commerce_gc_client_webhook()
 */
function commerce_gc_client_webhook_mandates($order, $action, array $event) {
  $log = $event['details']['description'];
  switch (TRUE) {
    case $action == 'created':
      commerce_order_status_update($order, 'processing', FALSE, TRUE, $log);
      break;

    case in_array($action, array(
      'submitted',
      'pending_customer_approval',
      'pending_submission',
      'customer_approval_granted',
      'customer_aproval_skipped',
      'transferred',
      'resubmission_requested',
    )):
      $order_wrapper = entity_metadata_wrapper('commerce_order', $order);
      $order->revision = TRUE;
      $order->log = $log;
      $order_wrapper->save();
      break;

    case in_array($action, array(
      'active',
      'reinstated',
      'replaced',
    )):
      commerce_order_status_update($order, 'completed', FALSE, TRUE, $log);
      break;

    case in_array($action, array(
      'failed',
      'cancelled',
      'expired',
    )):
      commerce_order_status_update($order, 'canceled', FALSE, TRUE, $log);
      break;

    default:
      $order_wrapper = entity_metadata_wrapper('commerce_order', $order);
      $order->revision = TRUE;
      $order->log = $log;
      $order_wrapper->save();
      break;
  }

  if ($action != 'created') {
    db_update('commerce_gc_client')->fields(array(
      'gc_mandate_status' => $action,
    ))->condition('order_id', $order->order_number)->execute();
  }
}

/**
 * Processes 'subscription' webhooks.
 *
 * @param object $order
 *   Commerce order entity.
 * @param string $action
 *   The GoCardless action that triggerd the sending of the webhook.
 * @param array $event
 *   The webhook event as provided by GoCardless.
 *
 * @see commerce_gc_client_webhook()
 */
function commerce_gc_client_webhook_subscriptions($order, $action, array $event) {

  if (!in_array($action, array('created', 'payment_created', 'amended'))) {
    db_update('commerce_gc_client_item')->fields(array(
      'gc_subscription_status' => $action,
    ))->condition('gc_subscription_id', $event['links']['subscription'])->execute();
  }
}

/**
 * Processes 'payments' webhooks.
 *
 * @param object $order
 *   Commerce order entity.
 * @param string $action
 *   The GoCardless action that triggerd the sending of the webhook.
 * @param array $event
 *   The webhook event as provided by GoCardless.
 * @param object $payment
 *   A GoCardless payment object that is retreived from via the API upon
 *   receipt of a 'payments' webhook.
 *
 * @see commerce_gc_client_webhook()
 */
function commerce_gc_client_webhook_payments($order, $action, array $event, $payment) {
  if ($action == 'created') {

    // Create transaction.
    $t = commerce_payment_transaction_new('gocardless_client', $order->order_id);
    $t->instance_id = $order->data['payment_method'];
    $t->amount = $payment->amount;
    $t->currency_code = $payment->currency;
    $t->message = $payment->description;
    $t->status = COMMERCE_PAYMENT_STATUS_PENDING;
    $t->remote_id = $payment->id;
    $t->remote_status = $payment->status;
  }
  else {

    $tid = db_select('commerce_payment_transaction', 't')
      ->fields('t', array('transaction_id'))
      ->condition('remote_id', $payment->id)
      ->execute()->fetchField();

    if (empty($tid)) {
      return;
    }

    $t = commerce_payment_transaction_load($tid);
    $t->message = $event['details']['description'];
    $t->remote_status = $action;

    switch (TRUE) {
      case in_array($action, array(
        'pending_customer_approval',
        'pending_submission',
        'submitted',
      )):
        $t->status = COMMERCE_PAYMENT_STATUS_PENDING;
        break;

      case in_array($action, array(
        'confirmed',
        'paid_out',
      )):
        $t->status = COMMERCE_PAYMENT_STATUS_SUCCESS;
        break;

      case in_array($action, array(
        'cancelled',
        'customer_approval_denied',
        'failed',
        'charged_back',
      )):
        $t->status = COMMERCE_PAYMENT_STATUS_FAILURE;
        break;
    }
  }
  commerce_payment_transaction_save($t);
}
